@@ -1,13 +1,13 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "org-node" "3.16.0"
+(define-package "org-node" "3.17.2"
   "Fast org-roam replacement."
   '((emacs         "29.1")
-    (llama         "0.5.0")
-    (org-mem       "0.29.2")
-    (magit-section "4.3.0"))
+    (llama         "1.0")
+    (magit-section "4.3.0")
+    (org-mem       "0.32.0"))
   :url "https://github.com/meedstrom/org-node"
-  :commit "d2dd24fc7264a5f621a4ed55fc4c6450714202df"
-  :revdesc "3.16.0-0-gd2dd24fc7264"
+  :commit "0ab480012fe0c40d2894546cdef909495d582a0d"
+  :revdesc "3.17.2-0-g0ab480012fe0"
   :keywords '("org" "hypermedia")
   :authors '(("Martin Edström" . "meedstrom91@gmail.com"))
   :maintainers '(("Martin Edström" . "meedstrom91@gmail.com")))
